@@ -1,18 +1,39 @@
 import os
 import platform
 import socket
+import sys
 import sysconfig
-from typing import List, Any, Tuple, Union
+from typing import List, Any, Tuple, Union, Optional
 import warnings
+import logging
+from threading import Lock
 
+from bugsnag.breadcrumbs import (
+    BreadcrumbType,
+    Breadcrumb,
+    Breadcrumbs,
+    OnBreadcrumbCallback
+)
 from bugsnag.sessiontracker import SessionMiddleware
-from bugsnag.middleware import DefaultMiddleware, MiddlewareStack
-from bugsnag.utils import (fully_qualified_class_name, validate_str_setter,
-                           validate_bool_setter, validate_iterable_setter,
-                           validate_required_str_setter)
+from bugsnag.middleware import (
+    DefaultMiddleware,
+    MiddlewareStack,
+    skip_bugsnag_middleware
+)
+from bugsnag.utils import (
+    fully_qualified_class_name,
+    partly_qualified_class_name,
+    validate_str_setter,
+    validate_bool_setter,
+    validate_iterable_setter,
+    validate_required_str_setter,
+    validate_int_setter,
+    validate_path_setter
+)
 from bugsnag.delivery import (create_default_delivery, DEFAULT_ENDPOINT,
                               DEFAULT_SESSIONS_ENDPOINT)
 from bugsnag.uwsgi import warn_if_running_uwsgi_without_threads
+from bugsnag.error import Error
 
 try:
     from contextvars import ContextVar
@@ -22,14 +43,26 @@ except ImportError:
     _request_info = ThreadContextVar('bugsnag-request', default=None)  # type: ignore  # noqa: E501
 
 
+try:
+    from os import PathLike
+except ImportError:
+    # PathLike was added in Python 3.6 so fallback to PurePath on Python 3.5 as
+    # all builtin Path objects inherit from PurePath
+    from pathlib import PurePath as PathLike  # type: ignore
+
+
 __all__ = ('Configuration', 'RequestConfiguration')
+_sentinel = object()
 
 
 class Configuration:
     """
     Global app-level Bugsnag configuration settings.
     """
-    def __init__(self):
+
+    def __init__(self, logger=_sentinel):
+        self._mutex = Lock()
+
         self.api_key = os.environ.get('BUGSNAG_API_KEY', None)
         self.release_stage = os.environ.get("BUGSNAG_RELEASE_STAGE",
                                             "production")
@@ -58,6 +91,7 @@ class Configuration:
         self.middleware = MiddlewareStack()
 
         self.internal_middleware = MiddlewareStack()
+        self.internal_middleware.before_notify(skip_bugsnag_middleware)
         self.internal_middleware.append(DefaultMiddleware)
         self.internal_middleware.append(SessionMiddleware)
 
@@ -70,6 +104,14 @@ class Configuration:
 
         self.runtime_versions = {"python": platform.python_version()}
 
+        self.logger = logger
+
+        self._max_breadcrumbs = 25
+        self.breadcrumb_log_level = logging.INFO
+        self.enabled_breadcrumb_types = list(BreadcrumbType)
+        self._breadcrumbs = Breadcrumbs(self.max_breadcrumbs)
+        self._on_breadcrumbs = []
+
     def configure(self, api_key=None, app_type=None, app_version=None,
                   asynchronous=None, auto_notify=None,
                   auto_capture_sessions=None, delivery=None, endpoint=None,
@@ -77,7 +119,9 @@ class Configuration:
                   notify_release_stages=None, params_filters=None,
                   project_root=None, proxy_host=None, release_stage=None,
                   send_code=None, send_environment=None, session_endpoint=None,
-                  traceback_exclude_modules=None):
+                  traceback_exclude_modules=None, logger=_sentinel,
+                  breadcrumb_log_level=None, enabled_breadcrumb_types=None,
+                  max_breadcrumbs=None):
         """
         Validate and set configuration options. Will warn if an option is of an
         incorrect type.
@@ -122,6 +166,15 @@ class Configuration:
             self.session_endpoint = session_endpoint
         if traceback_exclude_modules is not None:
             self.traceback_exclude_modules = traceback_exclude_modules
+        if logger is not _sentinel:
+            self.logger = logger
+        if breadcrumb_log_level is not None:
+            self.breadcrumb_log_level = breadcrumb_log_level
+        if enabled_breadcrumb_types is not None:
+            self.enabled_breadcrumb_types = enabled_breadcrumb_types
+        if max_breadcrumbs is not None:
+            self.max_breadcrumbs = max_breadcrumbs
+
         return self
 
     def get(self, name):
@@ -324,9 +377,9 @@ class Configuration:
         return self._project_root
 
     @project_root.setter  # type: ignore
-    @validate_str_setter
-    def project_root(self, value: str):
-        self._project_root = value
+    @validate_path_setter
+    def project_root(self, value: Union[str, PathLike]):
+        self._project_root = str(value)
 
     @property
     def proxy_host(self):
@@ -407,14 +460,106 @@ class Configuration:
     def traceback_exclude_modules(self, value: List[str]):
         self._traceback_exclude_modules = value
 
+    @property
+    def logger(self) -> logging.Logger:
+        """
+        Logger for use internally
+        """
+        return self._logger
+
+    @logger.setter  # type: ignore
+    def logger(self, logger: Optional[logging.Logger]) -> None:
+        if logger is _sentinel:
+            logger = self._create_default_logger()
+        elif logger is None:
+            logger = self._create_null_logger()
+
+        if not isinstance(logger, logging.Logger):
+            actual = type(logger).__name__
+            message = 'logger should be logging.Logger, got ' + actual
+            warnings.warn(message, RuntimeWarning)
+
+            logger = self._create_default_logger()
+
+        self._logger = logger
+
+    @property
+    def max_breadcrumbs(self) -> int:
+        return self._max_breadcrumbs
+
+    @max_breadcrumbs.setter  # type: ignore
+    @validate_int_setter
+    def max_breadcrumbs(self, new_max: int) -> None:
+        if 0 <= new_max <= 100:
+            self._breadcrumbs.resize(new_max)
+            self._max_breadcrumbs = new_max
+        else:
+            message = (
+                'max_breadcrumbs should be an int between 0 and 100, got "{}"'
+            ).format(new_max)
+
+            warnings.warn(message, RuntimeWarning)
+
+    @property
+    def breadcrumbs(self) -> List[Breadcrumb]:
+        return self._breadcrumbs.to_list()
+
+    def add_on_breadcrumb(self, on_breadcrumb: OnBreadcrumbCallback) -> None:
+        with self._mutex:
+            self._on_breadcrumbs.append(on_breadcrumb)
+
+    def remove_on_breadcrumb(
+        self,
+        on_breadcrumb: OnBreadcrumbCallback
+    ) -> None:
+        with self._mutex:
+            try:
+                self._on_breadcrumbs.remove(on_breadcrumb)
+            except ValueError:
+                # ignore exception if "on_breadcrumb" is not in the list
+                pass
+
     def should_notify(self) -> bool:
         return self.notify_release_stages is None or \
             (isinstance(self.notify_release_stages, (tuple, list)) and
              self.release_stage in self.notify_release_stages)
 
-    def should_ignore(self, exception: BaseException) -> bool:
-        return self.ignore_classes is not None and \
-            fully_qualified_class_name(exception) in self.ignore_classes
+    def should_ignore(
+        self,
+        exception: Union[BaseException, List[Error]]
+    ) -> bool:
+        if self.ignore_classes is None:
+            return False
+
+        # if this is a list of Error instances, check if any of the error
+        # classes are ignored
+        if isinstance(exception, list):
+            return any(e.error_class in self.ignore_classes for e in exception)
+
+        return (fully_qualified_class_name(exception) in self.ignore_classes or
+                partly_qualified_class_name(exception) in self.ignore_classes)
+
+    def _create_default_logger(self) -> logging.Logger:
+        logger = logging.getLogger('bugsnag')
+        logger.setLevel(logging.WARNING)
+
+        format = '%(asctime)s - [%(name)s] %(levelname)s - %(message)s'
+        formatter = logging.Formatter(format)
+
+        handler = logging.StreamHandler(sys.stderr)
+        handler.setFormatter(formatter)
+
+        logger.addHandler(handler)
+
+        return logger
+
+    def _create_null_logger(self) -> logging.Logger:
+        logger = logging.getLogger('bugsnag')
+        logger.handlers = []
+
+        logger.addHandler(logging.NullHandler())
+
+        return logger
 
 
 class RequestConfiguration:
